/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.datamodel.AlignmentI;
import jalview.datamodel.HiddenColumns;
import jalview.datamodel.HiddenColumns.VisibleBlocksVisBoundsIterator;
import jalview.datamodel.SearchResultsI;
import jalview.datamodel.SequenceGroup;
import jalview.datamodel.SequenceI;
import jalview.renderer.ScaleRenderer;
import jalview.renderer.ScaleRenderer.ScaleMark;
import jalview.viewmodel.ViewportListenerI;
import jalview.viewmodel.ViewportRanges;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.util.Iterator;

import javax.swing.JComponent;

/**
 * DOCUMENT ME!
 * 
 * @author $author$
 * @version $Revision$
 */
public class SeqCanvas extends JComponent implements ViewportListenerI
{
  private static String ZEROS = "0000000000";

  final FeatureRenderer fr;

  final SequenceRenderer seqRdr;

  BufferedImage img;

  Graphics2D gg;

  AlignViewport av;

  boolean fastPaint = false;

  int labelWidthWest;

  int labelWidthEast;

  int cursorX = 0;

  int cursorY = 0;

  int charHeight = 0;

  int charWidth = 0;

  boolean fastpainting = false;

  AnnotationPanel annotations;

  /**
   * Creates a new SeqCanvas object.
   * 
   * @param av
   *          DOCUMENT ME!
   */
  public SeqCanvas(AlignmentPanel ap)
  {
    this.av = ap.av;
    updateViewport();
    fr = new FeatureRenderer(ap);
    seqRdr = new SequenceRenderer(av);
    setLayout(new BorderLayout());
    PaintRefresher.Register(this, av.getSequenceSetId());
    setBackground(Color.white);

    av.getRanges().addPropertyChangeListener(this);
  }

  public SequenceRenderer getSequenceRenderer()
  {
    return seqRdr;
  }

  public FeatureRenderer getFeatureRenderer()
  {
    return fr;
  }

  private void updateViewport()
  {
    charHeight = av.getCharHeight();
    charWidth = av.getCharWidth();
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param startx
   *          DOCUMENT ME!
   * @param endx
   *          DOCUMENT ME!
   * @param ypos
   *          DOCUMENT ME!
   */
  private void drawNorthScale(Graphics g, int startx, int endx, int ypos)
  {
    updateViewport();
    for (ScaleMark mark : new ScaleRenderer().calculateMarks(av, startx,
            endx))
    {
      int mpos = mark.column; // (i - startx - 1)
      if (mpos < 0)
      {
        continue;
      }
      String mstring = mark.text;

      if (mark.major)
      {
        if (mstring != null)
        {
          g.drawString(mstring, mpos * charWidth, ypos - (charHeight / 2));
        }
        g.drawLine((mpos * charWidth) + (charWidth / 2),
                (ypos + 2) - (charHeight / 2),
                (mpos * charWidth) + (charWidth / 2), ypos - 2);
      }
    }
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param startx
   *          DOCUMENT ME!
   * @param endx
   *          DOCUMENT ME!
   * @param ypos
   *          DOCUMENT ME!
   */
  void drawWestScale(Graphics g, int startx, int endx, int ypos)
  {
    FontMetrics fm = getFontMetrics(av.getFont());
    ypos += charHeight;

    if (av.hasHiddenColumns())
    {
      startx = av.getAlignment().getHiddenColumns()
              .adjustForHiddenColumns(startx);
      endx = av.getAlignment().getHiddenColumns()
              .adjustForHiddenColumns(endx);
    }

    int maxwidth = av.getAlignment().getWidth();
    if (av.hasHiddenColumns())
    {
      maxwidth = av.getAlignment().getHiddenColumns()
              .findColumnPosition(maxwidth) - 1;
    }

    // WEST SCALE
    for (int i = 0; i < av.getAlignment().getHeight(); i++)
    {
      SequenceI seq = av.getAlignment().getSequenceAt(i);
      int index = startx;
      int value = -1;

      while (index < endx)
      {
        if (jalview.util.Comparison.isGap(seq.getCharAt(index)))
        {
          index++;

          continue;
        }

        value = av.getAlignment().getSequenceAt(i).findPosition(index);

        break;
      }

      if (value != -1)
      {
        int x = labelWidthWest - fm.stringWidth(String.valueOf(value))
                - charWidth / 2;
        g.drawString(value + "", x,
                (ypos + (i * charHeight)) - (charHeight / 5));
      }
    }
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param startx
   *          DOCUMENT ME!
   * @param endx
   *          DOCUMENT ME!
   * @param ypos
   *          DOCUMENT ME!
   */
  void drawEastScale(Graphics g, int startx, int endx, int ypos)
  {
    ypos += charHeight;

    if (av.hasHiddenColumns())
    {
      endx = av.getAlignment().getHiddenColumns()
              .adjustForHiddenColumns(endx);
    }

    SequenceI seq;
    // EAST SCALE
    for (int i = 0; i < av.getAlignment().getHeight(); i++)
    {
      seq = av.getAlignment().getSequenceAt(i);
      int index = endx;
      int value = -1;

      while (index > startx)
      {
        if (jalview.util.Comparison.isGap(seq.getCharAt(index)))
        {
          index--;

          continue;
        }

        value = seq.findPosition(index);

        break;
      }

      if (value != -1)
      {
        g.drawString(String.valueOf(value), 0,
                (ypos + (i * charHeight)) - (charHeight / 5));
      }
    }
  }


  /**
   * need to make this thread safe move alignment rendering in response to
   * slider adjustment
   * 
   * @param horizontal
   *          shift along
   * @param vertical
   *          shift up or down in repaint
   */
  public void fastPaint(int horizontal, int vertical)
  {
    if (fastpainting || gg == null || img == null)
    {
      return;
    }
    fastpainting = true;
    fastPaint = true;
    updateViewport();

    ViewportRanges ranges = av.getRanges();
    int startRes = ranges.getStartRes();
    int endRes = ranges.getEndRes();
    int startSeq = ranges.getStartSeq();
    int endSeq = ranges.getEndSeq();
    int transX = 0;
    int transY = 0;

    gg.copyArea(horizontal * charWidth, vertical * charHeight,
            img.getWidth(), img.getHeight(), -horizontal * charWidth,
            -vertical * charHeight);

    if (horizontal > 0) // scrollbar pulled right, image to the left
    {
      transX = (endRes - startRes - horizontal) * charWidth;
      startRes = endRes - horizontal;
    }
    else if (horizontal < 0)
    {
      endRes = startRes - horizontal;
    }
    else if (vertical > 0) // scroll down
    {
      startSeq = endSeq - vertical;

      if (startSeq < ranges.getStartSeq())
      { // ie scrolling too fast, more than a page at a time
        startSeq = ranges.getStartSeq();
      }
      else
      {
        transY = img.getHeight() - ((vertical + 1) * charHeight);
      }
    }
    else if (vertical < 0)
    {
      endSeq = startSeq - vertical;

      if (endSeq > ranges.getEndSeq())
      {
        endSeq = ranges.getEndSeq();
      }
    }

    gg.translate(transX, transY);
    drawPanel(gg, startRes, endRes, startSeq, endSeq, 0);
    gg.translate(-transX, -transY);

    repaint();
    fastpainting = false;
  }

  @Override
  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);

    updateViewport();

    ViewportRanges ranges = av.getRanges();

    int width = getWidth();
    int height = getHeight();

    width -= (width % charWidth);
    height -= (height % charHeight);

    // selectImage is the selection group outline image
    BufferedImage selectImage = drawSelectionGroup(
            ranges.getStartRes(), ranges.getEndRes(),
            ranges.getStartSeq(), ranges.getEndSeq());

    if ((img != null) && (fastPaint
            || (getVisibleRect().width != g.getClipBounds().width)
            || (getVisibleRect().height != g.getClipBounds().height)))
    {
      BufferedImage lcimg = buildLocalImage(selectImage);
      g.drawImage(lcimg, 0, 0, this);
      fastPaint = false;
    }
    else if ((width > 0) && (height > 0))
    {
      // img is a cached version of the last view we drew, if any
      // if we have no img or the size has changed, make a new one
      if (img == null || width != img.getWidth()
              || height != img.getHeight())
      {
        img = setupImage();
        if (img == null)
        {
          return;
        }
        gg = (Graphics2D) img.getGraphics();
        gg.setFont(av.getFont());
      }

      if (av.antiAlias)
      {
        gg.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);
      }

      gg.setColor(Color.white);
      gg.fillRect(0, 0, img.getWidth(), img.getHeight());

      if (av.getWrapAlignment())
      {
        drawWrappedPanel(gg, getWidth(), getHeight(), ranges.getStartRes());
      }
      else
      {
        drawPanel(gg, ranges.getStartRes(), ranges.getEndRes(),
                ranges.getStartSeq(), ranges.getEndSeq(), 0);
      }

      // lcimg is a local *copy* of img which we'll draw selectImage on top of
      BufferedImage lcimg = buildLocalImage(selectImage);
      g.drawImage(lcimg, 0, 0, this);
    }
  }

  /**
   * Draw an alignment panel for printing
   * 
   * @param g1
   *          Graphics object to draw with
   * @param startRes
   *          start residue of print area
   * @param endRes
   *          end residue of print area
   * @param startSeq
   *          start sequence of print area
   * @param endSeq
   *          end sequence of print area
   */
  public void drawPanelForPrinting(Graphics g1, int startRes, int endRes,
          int startSeq, int endSeq)
  {
    drawPanel(g1, startRes, endRes, startSeq, endSeq, 0);

    BufferedImage selectImage = drawSelectionGroup(startRes, endRes,
            startSeq, endSeq);
    if (selectImage != null)
    {
      ((Graphics2D) g1).setComposite(AlphaComposite
              .getInstance(AlphaComposite.SRC_OVER));
      g1.drawImage(selectImage, 0, 0, this);
    }
  }

  /**
   * Draw a wrapped alignment panel for printing
   * 
   * @param g
   *          Graphics object to draw with
   * @param canvasWidth
   *          width of drawing area
   * @param canvasHeight
   *          height of drawing area
   * @param startRes
   *          start residue of print area
   */
  public void drawWrappedPanelForPrinting(Graphics g, int canvasWidth,
          int canvasHeight, int startRes)
  {
    SequenceGroup group = av.getSelectionGroup();

    drawWrappedPanel(g, canvasWidth, canvasHeight, startRes);

    if (group != null)
    {
      BufferedImage selectImage = null;
      try
      {
        selectImage = new BufferedImage(canvasWidth, canvasHeight,
                BufferedImage.TYPE_INT_ARGB); // ARGB so alpha compositing works
      } catch (OutOfMemoryError er)
      {
        System.gc();
        System.err.println("Print image OutOfMemory Error.\n" + er);
        new OOMWarning("Creating wrapped alignment image for printing", er);
      }
      if (selectImage != null)
      {
        Graphics2D g2 = selectImage.createGraphics();
        setupSelectionGroup(g2, selectImage);
        drawWrappedSelection(g2, group, canvasWidth, canvasHeight,
                startRes);

        g2.setComposite(
                AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
        g.drawImage(selectImage, 0, 0, this);
        g2.dispose();
      }
    }
  }

  /*
   * Make a local image by combining the cached image img
   * with any selection
   */
  private BufferedImage buildLocalImage(BufferedImage selectImage)
  {
    // clone the cached image
    BufferedImage lcimg = new BufferedImage(img.getWidth(), img.getHeight(),
            img.getType());
    Graphics2D g2d = lcimg.createGraphics();
    g2d.drawImage(img, 0, 0, null);

    // overlay selection group on lcimg
    if (selectImage != null)
    {
      g2d.setComposite(
              AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
      g2d.drawImage(selectImage, 0, 0, this);
    }
    g2d.dispose();

    return lcimg;
  }

  /*
   * Set up a buffered image of the correct height and size for the sequence canvas
   */
  private BufferedImage setupImage()
  {
    BufferedImage lcimg = null;

    int width = getWidth();
    int height = getHeight();

    width -= (width % charWidth);
    height -= (height % charHeight);

    if ((width < 1) || (height < 1))
    {
      return null;
    }

    try
    {
      lcimg = new BufferedImage(width, height,
              BufferedImage.TYPE_INT_ARGB); // ARGB so alpha compositing works
    } catch (OutOfMemoryError er)
    {
      System.gc();
      System.err.println(
              "Group image OutOfMemory Redraw Error.\n" + er);
      new OOMWarning("Creating alignment image for display", er);

      return null;
    }

    return lcimg;
  }

  /**
   * Returns the visible width of the canvas in residues, after allowing for
   * East or West scales (if shown)
   * 
   * @param canvasWidth
   *          the width in pixels (possibly including scales)
   * 
   * @return
   */
  public int getWrappedCanvasWidth(int canvasWidth)
  {
    FontMetrics fm = getFontMetrics(av.getFont());

    labelWidthEast = 0;
    labelWidthWest = 0;

    if (av.getScaleRightWrapped())
    {
      labelWidthEast = getLabelWidth(fm);
    }

    if (av.getScaleLeftWrapped())
    {
      labelWidthWest = labelWidthEast > 0 ? labelWidthEast
              : getLabelWidth(fm);
    }

    return (canvasWidth - labelWidthEast - labelWidthWest) / charWidth;
  }

  /**
   * Returns a pixel width suitable for showing the largest sequence coordinate
   * (end position) in the alignment. Returns 2 plus the number of decimal
   * digits to be shown (3 for 1-10, 4 for 11-99 etc).
   * 
   * @param fm
   * @return
   */
  protected int getLabelWidth(FontMetrics fm)
  {
    /*
     * find the biggest sequence end position we need to show
     * (note this is not necessarily the sequence length)
     */
    int maxWidth = 0;
    AlignmentI alignment = av.getAlignment();
    for (int i = 0; i < alignment.getHeight(); i++)
    {
      maxWidth = Math.max(maxWidth, alignment.getSequenceAt(i).getEnd());
    }

    int length = 2;
    for (int i = maxWidth; i > 0; i /= 10)
    {
      length++;
    }

    return fm.stringWidth(ZEROS.substring(0, length));
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param canvasWidth
   *          DOCUMENT ME!
   * @param canvasHeight
   *          DOCUMENT ME!
   * @param startRes
   *          DOCUMENT ME!
   */
  private void drawWrappedPanel(Graphics g, int canvasWidth,
          int canvasHeight, int startRes)
  {
    updateViewport();
    AlignmentI al = av.getAlignment();

    int labelWidth = 0;
    if (av.getScaleRightWrapped() || av.getScaleLeftWrapped())
    {
      FontMetrics fm = getFontMetrics(av.getFont());
      labelWidth = getLabelWidth(fm);
    }

    labelWidthEast = av.getScaleRightWrapped() ? labelWidth : 0;
    labelWidthWest = av.getScaleLeftWrapped() ? labelWidth : 0;

    int hgap = charHeight;
    if (av.getScaleAboveWrapped())
    {
      hgap += charHeight;
    }

    int cWidth = (canvasWidth - labelWidthEast - labelWidthWest) / charWidth;
    int cHeight = av.getAlignment().getHeight() * charHeight;

    av.setWrappedWidth(cWidth);

    av.getRanges().setViewportStartAndWidth(startRes, cWidth);

    int endx;
    int ypos = hgap;
    int maxwidth = av.getAlignment().getWidth();

    if (av.hasHiddenColumns())
    {
      maxwidth = av.getAlignment().getHiddenColumns()
              .findColumnPosition(maxwidth);
    }

    int annotationHeight = getAnnotationHeight();

    while ((ypos <= canvasHeight) && (startRes < maxwidth))
    {
      endx = startRes + cWidth - 1;

      if (endx > maxwidth)
      {
        endx = maxwidth;
      }

      g.setFont(av.getFont());
      g.setColor(Color.black);

      if (av.getScaleLeftWrapped())
      {
        drawWestScale(g, startRes, endx, ypos);
      }

      if (av.getScaleRightWrapped())
      {
        g.translate(canvasWidth - labelWidthEast, 0);
        drawEastScale(g, startRes, endx, ypos);
        g.translate(-(canvasWidth - labelWidthEast), 0);
      }

      g.translate(labelWidthWest, 0);

      if (av.getScaleAboveWrapped())
      {
        drawNorthScale(g, startRes, endx, ypos);
      }

      if (av.hasHiddenColumns() && av.getShowHiddenMarkers())
      {
        g.setColor(Color.blue);
        int res;
        HiddenColumns hidden = av.getAlignment().getHiddenColumns();

        Iterator<Integer> it = hidden.getBoundedStartIterator(startRes,
                endx + 1);
        while (it.hasNext())
        {
          res = it.next() - startRes;
          gg.fillPolygon(
                  new int[]
          { res * charWidth - charHeight / 4,
              res * charWidth + charHeight / 4, res * charWidth },
                  new int[]
          { ypos - (charHeight / 2), ypos - (charHeight / 2),
              ypos - (charHeight / 2) + 8 }, 3);
        }
      }

      // When printing we have an extra clipped region,
      // the Printable page which we need to account for here
      Shape clip = g.getClip();

      if (clip == null)
      {
        g.setClip(0, 0, cWidth * charWidth, canvasHeight);
      }
      else
      {
        g.setClip(0, (int) clip.getBounds().getY(), cWidth * charWidth,
                (int) clip.getBounds().getHeight());
      }

      drawPanel(g, startRes, endx, 0, al.getHeight() - 1, ypos);

      if (av.isShowAnnotation())
      {
        g.translate(0, cHeight + ypos + 3);
        if (annotations == null)
        {
          annotations = new AnnotationPanel(av);
        }

        annotations.renderer.drawComponent(annotations, av, g, -1, startRes,
                endx + 1);
        g.translate(0, -cHeight - ypos - 3);
      }
      g.setClip(clip);
      g.translate(-labelWidthWest, 0);

      ypos += cHeight + annotationHeight + hgap;

      startRes += cWidth;
    }
  }

  /*
   * Draw a selection group over a wrapped alignment
   */
  private void drawWrappedSelection(Graphics2D g, SequenceGroup group,
          int canvasWidth,
          int canvasHeight, int startRes)
  {
    // height gap above each panel
    int hgap = charHeight;
    if (av.getScaleAboveWrapped())
    {
      hgap += charHeight;
    }

    int cWidth = (canvasWidth - labelWidthEast - labelWidthWest)
            / charWidth;
    int cHeight = av.getAlignment().getHeight() * charHeight;

    int startx = startRes;
    int endx;
    int ypos = hgap; // vertical offset
    int maxwidth = av.getAlignment().getWidth();

    if (av.hasHiddenColumns())
    {
      maxwidth = av.getAlignment().getHiddenColumns()
              .findColumnPosition(maxwidth);
    }

    // chop the wrapped alignment extent up into panel-sized blocks and treat
    // each block as if it were a block from an unwrapped alignment
    while ((ypos <= canvasHeight) && (startx < maxwidth))
    {
      // set end value to be start + width, or maxwidth, whichever is smaller
      endx = startx + cWidth - 1;

      if (endx > maxwidth)
      {
        endx = maxwidth;
      }

      g.translate(labelWidthWest, 0);

      drawUnwrappedSelection(g, group, startx, endx, 0,
              av.getAlignment().getHeight() - 1,
              ypos);

      g.translate(-labelWidthWest, 0);

      // update vertical offset
      ypos += cHeight + getAnnotationHeight() + hgap;

      // update horizontal offset
      startx += cWidth;
    }
  }

  int getAnnotationHeight()
  {
    if (!av.isShowAnnotation())
    {
      return 0;
    }

    if (annotations == null)
    {
      annotations = new AnnotationPanel(av);
    }

    return annotations.adjustPanelHeight();
  }

  /**
   * Draws the visible region of the alignment on the graphics context. If there
   * are hidden column markers in the visible region, then each sub-region
   * between the markers is drawn separately, followed by the hidden column
   * marker.
   * 
   * @param g1
   *          Graphics object to draw with
   * @param startRes
   *          offset of the first column in the visible region (0..)
   * @param endRes
   *          offset of the last column in the visible region (0..)
   * @param startSeq
   *          offset of the first sequence in the visible region (0..)
   * @param endSeq
   *          offset of the last sequence in the visible region (0..)
   * @param yOffset
   *          vertical offset at which to draw (for wrapped alignments)
   */
  public void drawPanel(Graphics g1, final int startRes, final int endRes,
          final int startSeq, final int endSeq, final int yOffset)
  {
    updateViewport();
    if (!av.hasHiddenColumns())
    {
      draw(g1, startRes, endRes, startSeq, endSeq, yOffset);
    }
    else
    {
      int screenY = 0;
      int blockStart;
      int blockEnd;

      HiddenColumns hidden = av.getAlignment().getHiddenColumns();
      VisibleBlocksVisBoundsIterator regions = (VisibleBlocksVisBoundsIterator) hidden
              .getVisibleBlocksIterator(startRes, endRes, true);

      while (regions.hasNext())
      {
        int[] region = regions.next();
        blockEnd = region[1];
        blockStart = region[0];

        /*
         * draw up to just before the next hidden region, or the end of
         * the visible region, whichever comes first
         */
        g1.translate(screenY * charWidth, 0);

        draw(g1, blockStart, blockEnd, startSeq, endSeq, yOffset);

        /*
         * draw the downline of the hidden column marker (ScalePanel draws the
         * triangle on top) if we reached it
         */
        if (av.getShowHiddenMarkers()
                && (regions.hasNext() || regions.endsAtHidden()))
        {
          g1.setColor(Color.blue);

          g1.drawLine((blockEnd - blockStart + 1) * charWidth - 1,
                  0 + yOffset, (blockEnd - blockStart + 1) * charWidth - 1,
                  (endSeq - startSeq + 1) * charHeight + yOffset);
        }

        g1.translate(-screenY * charWidth, 0);
        screenY += blockEnd - blockStart + 1;
      }
    }

  }

  /**
   * Draws a region of the visible alignment
   * 
   * @param g1
   * @param startRes
   *          offset of the first column in the visible region (0..)
   * @param endRes
   *          offset of the last column in the visible region (0..)
   * @param startSeq
   *          offset of the first sequence in the visible region (0..)
   * @param endSeq
   *          offset of the last sequence in the visible region (0..)
   * @param yOffset
   *          vertical offset at which to draw (for wrapped alignments)
   */
  private void draw(Graphics g, int startRes, int endRes, int startSeq,
          int endSeq, int offset)
  {
    g.setFont(av.getFont());
    seqRdr.prepare(g, av.isRenderGaps());

    SequenceI nextSeq;

    // / First draw the sequences
    // ///////////////////////////
    for (int i = startSeq; i <= endSeq; i++)
    {
      nextSeq = av.getAlignment().getSequenceAt(i);
      if (nextSeq == null)
      {
        // occasionally, a race condition occurs such that the alignment row is
        // empty
        continue;
      }
      seqRdr.drawSequence(nextSeq, av.getAlignment().findAllGroups(nextSeq),
              startRes, endRes, offset + ((i - startSeq) * charHeight));

      if (av.isShowSequenceFeatures())
      {
        fr.drawSequence(g, nextSeq, startRes, endRes,
                offset + ((i - startSeq) * charHeight), false);
      }

      /*
       * highlight search Results once sequence has been drawn
       */
      if (av.hasSearchResults())
      {
        SearchResultsI searchResults = av.getSearchResults();
        int[] visibleResults = searchResults.getResults(nextSeq,
                startRes, endRes);
        if (visibleResults != null)
        {
          for (int r = 0; r < visibleResults.length; r += 2)
          {
            seqRdr.drawHighlightedText(nextSeq, visibleResults[r],
                    visibleResults[r + 1], (visibleResults[r] - startRes)
                            * charWidth, offset
                            + ((i - startSeq) * charHeight));
          }
        }
      }

      if (av.cursorMode && cursorY == i && cursorX >= startRes
              && cursorX <= endRes)
      {
        seqRdr.drawCursor(nextSeq, cursorX, (cursorX - startRes) * charWidth,
                offset + ((i - startSeq) * charHeight));
      }
    }

    if (av.getSelectionGroup() != null
            || av.getAlignment().getGroups().size() > 0)
    {
      drawGroupsBoundaries(g, startRes, endRes, startSeq, endSeq, offset);
    }

  }

  void drawGroupsBoundaries(Graphics g1, int startRes, int endRes,
          int startSeq, int endSeq, int offset)
  {
    Graphics2D g = (Graphics2D) g1;
    //
    // ///////////////////////////////////
    // Now outline any areas if necessary
    // ///////////////////////////////////

    SequenceGroup group = null;
    int groupIndex = -1;

    if (av.getAlignment().getGroups().size() > 0)
    {
      group = av.getAlignment().getGroups().get(0);
      groupIndex = 0;
    }

    if (group != null)
    {
      g.setStroke(new BasicStroke());
      g.setColor(group.getOutlineColour());
      
      do
      {
        drawPartialGroupOutline(g, group, startRes, endRes, startSeq,
                endSeq, offset);

        groupIndex++;

        g.setStroke(new BasicStroke());

        if (groupIndex >= av.getAlignment().getGroups().size())
        {
          break;
        }

        group = av.getAlignment().getGroups().get(groupIndex);

      } while (groupIndex < av.getAlignment().getGroups().size());

    }

  }


  /*
   * Draw the selection group as a separate image and overlay
   */
  private BufferedImage drawSelectionGroup(int startRes, int endRes,
          int startSeq, int endSeq)
  {
    // get a new image of the correct size
    BufferedImage selectionImage = setupImage();

    if (selectionImage == null)
    {
      return null;
    }

    SequenceGroup group = av.getSelectionGroup();
    if (group == null)
    {
      // nothing to draw
      return null;
    }

    // set up drawing colour
    Graphics2D g = (Graphics2D) selectionImage.getGraphics();

    setupSelectionGroup(g, selectionImage);

    if (!av.getWrapAlignment())
    {
      drawUnwrappedSelection(g, group, startRes, endRes, startSeq, endSeq,
              0);
    }
    else
    {
      drawWrappedSelection(g, group, getWidth(), getHeight(),
              av.getRanges().getStartRes());
    }

    g.dispose();
    return selectionImage;
  }

  /*
   * Set up graphics for selection group
   */
  private void setupSelectionGroup(Graphics2D g,
          BufferedImage selectionImage)
  {
    // set background to transparent
    g.setComposite(AlphaComposite.getInstance(AlphaComposite.CLEAR, 0.0f));
    g.fillRect(0, 0, selectionImage.getWidth(), selectionImage.getHeight());

    // set up foreground to draw red dashed line
    g.setComposite(AlphaComposite.Src);
    g.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
            BasicStroke.JOIN_ROUND, 3f, new float[]
    { 5f, 3f }, 0f));
    g.setColor(Color.RED);
  }

  /*
   * Draw a selection group over an unwrapped alignment
   * @param g graphics object to draw with
   * @param group selection group
   * @param startRes start residue of area to draw
   * @param endRes end residue of area to draw
   * @param startSeq start sequence of area to draw
   * @param endSeq end sequence of area to draw
   * @param offset vertical offset (used when called from wrapped alignment code)
   */
  private void drawUnwrappedSelection(Graphics2D g, SequenceGroup group,
          int startRes, int endRes, int startSeq, int endSeq, int offset)
  {
    if (!av.hasHiddenColumns())
    {
      drawPartialGroupOutline(g, group, startRes, endRes, startSeq, endSeq,
              offset);
    }
    else
    {
      // package into blocks of visible columns
      int screenY = 0;
      int blockStart;
      int blockEnd;

      HiddenColumns hidden = av.getAlignment().getHiddenColumns();
      VisibleBlocksVisBoundsIterator regions = (VisibleBlocksVisBoundsIterator) hidden
              .getVisibleBlocksIterator(startRes, endRes, true);
      while (regions.hasNext())
      {
        int[] region = regions.next();
        blockEnd = region[1];
        blockStart = region[0];

        g.translate(screenY * charWidth, 0);
        drawPartialGroupOutline(g, group,
                blockStart, blockEnd, startSeq, endSeq, offset);

        g.translate(-screenY * charWidth, 0);
        screenY += blockEnd - blockStart + 1;
      }
    }
  }

  /*
   * Draw the selection group as a separate image and overlay
   */
  private void drawPartialGroupOutline(Graphics2D g, SequenceGroup group,
          int startRes, int endRes, int startSeq, int endSeq,
          int verticalOffset)
  {
    int visWidth = (endRes - startRes + 1) * charWidth;

    int oldY = -1;
    int i = 0;
    boolean inGroup = false;
    int top = -1;
    int bottom = -1;

    int sx = -1;
    int sy = -1;
    int xwidth = -1;

    for (i = startSeq; i <= endSeq; i++)
    {
      // position of start residue of group relative to startRes, in pixels
      sx = (group.getStartRes() - startRes) * charWidth;

      // width of group in pixels
      xwidth = (((group.getEndRes() + 1) - group.getStartRes()) * charWidth)
              - 1;

      sy = verticalOffset + (i - startSeq) * charHeight;

      if (sx + xwidth < 0 || sx > visWidth)
      {
        continue;
      }

      if ((sx <= (endRes - startRes) * charWidth)
              && group.getSequences(null)
                      .contains(av.getAlignment().getSequenceAt(i)))
      {
        if ((bottom == -1) && !group.getSequences(null)
                .contains(av.getAlignment().getSequenceAt(i + 1)))
        {
          bottom = sy + charHeight;
        }

        if (!inGroup)
        {
          if (((top == -1) && (i == 0)) || !group.getSequences(null)
                  .contains(av.getAlignment().getSequenceAt(i - 1)))
          {
            top = sy;
          }

          oldY = sy;
          inGroup = true;
        }
      }
      else
      {
        if (inGroup)
        {
          // if start position is visible, draw vertical line to left of
          // group
          if (sx >= 0 && sx < visWidth)
          {
            g.drawLine(sx, oldY, sx, sy);
          }

          // if end position is visible, draw vertical line to right of
          // group
          if (sx + xwidth < visWidth)
          {
            g.drawLine(sx + xwidth, oldY, sx + xwidth, sy);
          }

          if (sx < 0)
          {
            xwidth += sx;
            sx = 0;
          }

          // don't let width extend beyond current block, or group extent
          // fixes JAL-2672
          if (sx + xwidth >= (endRes - startRes + 1) * charWidth)
          {
            xwidth = (endRes - startRes + 1) * charWidth - sx;
          }
          
          // draw horizontal line at top of group
          if (top != -1)
          {
            g.drawLine(sx, top, sx + xwidth, top);
            top = -1;
          }

          // draw horizontal line at bottom of group
          if (bottom != -1)
          {
            g.drawLine(sx, bottom, sx + xwidth, bottom);
            bottom = -1;
          }

          inGroup = false;
        }
      }
    }

    if (inGroup)
    {
      sy = verticalOffset + ((i - startSeq) * charHeight);
      if (sx >= 0 && sx < visWidth)
      {
        g.drawLine(sx, oldY, sx, sy);
      }

      if (sx + xwidth < visWidth)
      {
        g.drawLine(sx + xwidth, oldY, sx + xwidth, sy);
      }

      if (sx < 0)
      {
        xwidth += sx;
        sx = 0;
      }

      if (sx + xwidth > visWidth)
      {
        xwidth = visWidth;
      }
      else if (sx + xwidth >= (endRes - startRes + 1) * charWidth)
      {
        xwidth = (endRes - startRes + 1) * charWidth;
      }

      if (top != -1)
      {
        g.drawLine(sx, top, sx + xwidth, top);
        top = -1;
      }

      if (bottom != -1)
      {
        g.drawLine(sx, bottom - 1, sx + xwidth, bottom - 1);
        bottom = -1;
      }

      inGroup = false;
    }
  }
  
  /**
   * Highlights search results in the visible region by rendering as white text
   * on a black background. Any previous highlighting is removed. Answers true
   * if any highlight was left on the visible alignment (so status bar should be
   * set to match), else false.
   * <p>
   * Currently fastPaint is not implemented for wrapped alignments. If a wrapped
   * alignment had to be scrolled to show the highlighted region, then it should
   * be fully redrawn, otherwise a fast paint can be performed. This argument
   * could be removed if fast paint of scrolled wrapped alignment is coded in
   * future (JAL-2609).
   * 
   * @param results
   * @param noFastPaint
   * @return
   */
  public boolean highlightSearchResults(SearchResultsI results,
          boolean noFastPaint)
  {
    if (fastpainting)
    {
      return false;
    }
    boolean wrapped = av.getWrapAlignment();

    try
    {
      fastPaint = !noFastPaint;
      fastpainting = fastPaint;

      updateViewport();

      /*
       * to avoid redrawing the whole visible region, we instead
       * redraw just the minimal regions to remove previous highlights
       * and add new ones
       */
      SearchResultsI previous = av.getSearchResults();
      av.setSearchResults(results);
      boolean redrawn = false;
      boolean drawn = false;
      if (wrapped)
      {
        redrawn = drawMappedPositionsWrapped(previous);
        drawn = drawMappedPositionsWrapped(results);
        redrawn |= drawn;
      }
      else
      {
        redrawn = drawMappedPositions(previous);
        drawn = drawMappedPositions(results);
        redrawn |= drawn;
      }

      /*
       * if highlights were either removed or added, repaint
       */
      if (redrawn)
      {
        repaint();
      }

      /*
       * return true only if highlights were added
       */
      return drawn;

    } finally
    {
      fastpainting = false;
    }
  }

  /**
   * Redraws the minimal rectangle in the visible region (if any) that includes
   * mapped positions of the given search results. Whether or not positions are
   * highlighted depends on the SearchResults set on the Viewport. This allows
   * this method to be called to either clear or set highlighting. Answers true
   * if any positions were drawn (in which case a repaint is still required),
   * else false.
   * 
   * @param results
   * @return
   */
  protected boolean drawMappedPositions(SearchResultsI results)
  {
    if (results == null)
    {
      return false;
    }

    /*
     * calculate the minimal rectangle to redraw that 
     * includes both new and existing search results
     */
    int firstSeq = Integer.MAX_VALUE;
    int lastSeq = -1;
    int firstCol = Integer.MAX_VALUE;
    int lastCol = -1;
    boolean matchFound = false;

    ViewportRanges ranges = av.getRanges();
    int firstVisibleColumn = ranges.getStartRes();
    int lastVisibleColumn = ranges.getEndRes();
    AlignmentI alignment = av.getAlignment();
    if (av.hasHiddenColumns())
    {
      firstVisibleColumn = alignment.getHiddenColumns()
              .adjustForHiddenColumns(firstVisibleColumn);
      lastVisibleColumn = alignment.getHiddenColumns()
              .adjustForHiddenColumns(lastVisibleColumn);
    }

    for (int seqNo = ranges.getStartSeq(); seqNo <= ranges
            .getEndSeq(); seqNo++)
    {
      SequenceI seq = alignment.getSequenceAt(seqNo);

      int[] visibleResults = results.getResults(seq, firstVisibleColumn,
              lastVisibleColumn);
      if (visibleResults != null)
      {
        for (int i = 0; i < visibleResults.length - 1; i += 2)
        {
          int firstMatchedColumn = visibleResults[i];
          int lastMatchedColumn = visibleResults[i + 1];
          if (firstMatchedColumn <= lastVisibleColumn
                  && lastMatchedColumn >= firstVisibleColumn)
          {
            /*
             * found a search results match in the visible region - 
             * remember the first and last sequence matched, and the first
             * and last visible columns in the matched positions
             */
            matchFound = true;
            firstSeq = Math.min(firstSeq, seqNo);
            lastSeq = Math.max(lastSeq, seqNo);
            firstMatchedColumn = Math.max(firstMatchedColumn,
                    firstVisibleColumn);
            lastMatchedColumn = Math.min(lastMatchedColumn,
                    lastVisibleColumn);
            firstCol = Math.min(firstCol, firstMatchedColumn);
            lastCol = Math.max(lastCol, lastMatchedColumn);
          }
        }
      }
    }

    if (matchFound)
    {
      if (av.hasHiddenColumns())
      {
        firstCol = alignment.getHiddenColumns()
                .findColumnPosition(firstCol);
        lastCol = alignment.getHiddenColumns().findColumnPosition(lastCol);
      }
      int transX = (firstCol - ranges.getStartRes()) * av.getCharWidth();
      int transY = (firstSeq - ranges.getStartSeq()) * av.getCharHeight();
      gg.translate(transX, transY);
      drawPanel(gg, firstCol, lastCol, firstSeq, lastSeq, 0);
      gg.translate(-transX, -transY);
    }

    return matchFound;
  }

  @Override
  public void propertyChange(PropertyChangeEvent evt)
  {
    String eventName = evt.getPropertyName();

    if (eventName.equals(SequenceGroup.SEQ_GROUP_CHANGED))
    {
      fastPaint = true;
      repaint();
    }
    else if (av.getWrapAlignment())
    {
      if (eventName.equals(ViewportRanges.STARTRES))
      {
        repaint();
      }
    }
    else
    {
      int scrollX = 0;
      if (eventName.equals(ViewportRanges.STARTRES))
      {
        // Make sure we're not trying to draw a panel
        // larger than the visible window
        ViewportRanges vpRanges = av.getRanges();
        scrollX = (int) evt.getNewValue() - (int) evt.getOldValue();
        int range = vpRanges.getEndRes() - vpRanges.getStartRes();
        if (scrollX > range)
        {
          scrollX = range;
        }
        else if (scrollX < -range)
        {
          scrollX = -range;
        }
      }

      // Both scrolling and resizing change viewport ranges: scrolling changes
      // both start and end points, but resize only changes end values.
      // Here we only want to fastpaint on a scroll, with resize using a normal
      // paint, so scroll events are identified as changes to the horizontal or
      // vertical start value.
      if (eventName.equals(ViewportRanges.STARTRES))
      {
        // scroll - startres and endres both change
        fastPaint(scrollX, 0);
      }
      else if (eventName.equals(ViewportRanges.STARTSEQ))
      {
        // scroll
        fastPaint(0, (int) evt.getNewValue() - (int) evt.getOldValue());
      }
    }
  }

  /**
   * Redraws any positions in the search results in the visible region of a
   * wrapped alignment. Any highlights are drawn depending on the search results
   * set on the Viewport, not the <code>results</code> argument. This allows
   * this method to be called either to clear highlights (passing the previous
   * search results), or to draw new highlights.
   * 
   * @param results
   * @return
   */
  protected boolean drawMappedPositionsWrapped(SearchResultsI results)
  {
    if (results == null)
    {
      return false;
    }
  
    boolean matchFound = false;

    int wrappedWidth = av.getWrappedWidth();
    int wrappedHeight = getRepeatHeightWrapped();

    ViewportRanges ranges = av.getRanges();
    int canvasHeight = getHeight();
    int repeats = canvasHeight / wrappedHeight;
    if (canvasHeight / wrappedHeight > 0)
    {
      repeats++;
    }

    int firstVisibleColumn = ranges.getStartRes();
    int lastVisibleColumn = ranges.getStartRes() + repeats
            * ranges.getViewportWidth() - 1;

    AlignmentI alignment = av.getAlignment();
    if (av.hasHiddenColumns())
    {
      firstVisibleColumn = alignment.getHiddenColumns()
              .adjustForHiddenColumns(firstVisibleColumn);
      lastVisibleColumn = alignment.getHiddenColumns()
              .adjustForHiddenColumns(lastVisibleColumn);
    }

    int gapHeight = charHeight * (av.getScaleAboveWrapped() ? 2 : 1);

    for (int seqNo = ranges.getStartSeq(); seqNo <= ranges
            .getEndSeq(); seqNo++)
    {
      SequenceI seq = alignment.getSequenceAt(seqNo);

      int[] visibleResults = results.getResults(seq, firstVisibleColumn,
              lastVisibleColumn);
      if (visibleResults != null)
      {
        for (int i = 0; i < visibleResults.length - 1; i += 2)
        {
          int firstMatchedColumn = visibleResults[i];
          int lastMatchedColumn = visibleResults[i + 1];
          if (firstMatchedColumn <= lastVisibleColumn
                  && lastMatchedColumn >= firstVisibleColumn)
          {
            /*
             * found a search results match in the visible region
             */
            firstMatchedColumn = Math.max(firstMatchedColumn,
                    firstVisibleColumn);
            lastMatchedColumn = Math.min(lastMatchedColumn,
                    lastVisibleColumn);

            /*
             * draw each mapped position separately (as contiguous positions may
             * wrap across lines)
             */
            for (int mappedPos = firstMatchedColumn; mappedPos <= lastMatchedColumn; mappedPos++)
            {
              int displayColumn = mappedPos;
              if (av.hasHiddenColumns())
              {
                displayColumn = alignment.getHiddenColumns()
                        .findColumnPosition(displayColumn);
              }

              /*
               * transX: offset from left edge of canvas to residue position
               */
              int transX = labelWidthWest
                      + ((displayColumn - ranges.getStartRes()) % wrappedWidth)
                      * av.getCharWidth();

              /*
               * transY: offset from top edge of canvas to residue position
               */
              int transY = gapHeight;
              transY += (displayColumn - ranges.getStartRes())
                      / wrappedWidth * wrappedHeight;
              transY += (seqNo - ranges.getStartSeq()) * av.getCharHeight();

              /*
               * yOffset is from graphics origin to start of visible region
               */
              int yOffset = 0;// (displayColumn / wrappedWidth) * wrappedHeight;
              if (transY < getHeight())
              {
                matchFound = true;
                gg.translate(transX, transY);
                drawPanel(gg, displayColumn, displayColumn, seqNo, seqNo,
                        yOffset);
                gg.translate(-transX, -transY);
              }
            }
          }
        }
      }
    }
  
    return matchFound;
  }

  /**
   * Answers the height in pixels of a repeating section of the wrapped
   * alignment, including space above, scale above if shown, sequences, and
   * annotation panel if shown
   * 
   * @return
   */
  protected int getRepeatHeightWrapped()
  {
    // gap (and maybe scale) above
    int repeatHeight = charHeight * (av.getScaleAboveWrapped() ? 2 : 1);

    // add sequences
    repeatHeight += av.getRanges().getViewportHeight() * charHeight;

    // add annotations panel height if shown
    repeatHeight += getAnnotationHeight();

    return repeatHeight;
  }
}
